/*
 * FTMEXT SDK example
 * Copyright (C) 2004 by IRCAM-Centre Georges Pompidou, Paris, France.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 * 
 * See file COPYING.LIB for further informations on licensing terms.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
#include "ftmlib.h"

typedef struct
{
  ftmext_t o;
  fts_array_t array;
  int size;
} pack_t;

/************************************************
 *
 *  user methods
 *
 */
FTMEXT_METHOD_VOID(pack, output)
{
  pack_t *self = (pack_t *)FTMEXT_GET_EXT();
  int size = fts_array_get_size(&self->array);
  fts_atom_t *atoms = fts_array_get_atoms(&self->array);
  
  ftmext_outlet_varargs((ftmext_t *)self, 0, size, atoms);
  
  FTMEXT_METHOD_RETURN;
}

FTMEXT_METHOD_VARARGS(pack, set)
{
  pack_t *self = (pack_t *)FTMEXT_GET_EXT();
  int ac = FTMEXT_GET_COUNT();
  fts_atom_t *at = FTMEXT_GET_ARGS();
  int size = fts_array_get_size(&self->array);  
  int i;
  
  if(ac > size)
	ac = size;
  
  for(i=0; i<ac; i++)
	fts_array_set_element(&self->array, i, at + i);	
  
  FTMEXT_METHOD_RETURN;
}

FTMEXT_INLET(pack)
{
  pack_t *self = (pack_t *)FTMEXT_GET_EXT();
  int inlet = FTMEXT_GET_INLET();
  fts_atom_t *a = FTMEXT_GET_ANY();
  int size = fts_array_get_size(&self->array);  
  
  if(inlet < size)
	fts_array_set_element(&self->array, inlet, a);	

  if(inlet == 0)
	pack_output((ftmext_t *)self);
    
  FTMEXT_INLET_RETURN;  
}

/* alternative implementation */
/*
FTMEXT_INLET(pack)
{
  pack_t *self = (pack_t *)FTMEXT_GET_EXT();
  int inlet = FTMEXT_GET_INLET();
  int ac = FTMEXT_GET_COUNT();
  fts_atom_t *at = FTMEXT_GET_ARGS();

  if(inlet + ac > size)
	ac = size - inlet;
	
  for(i=0; i<ac, i++)
	fts_array_set_element(&self->input, inlet + i, at + i);
  
  if(inlet == 0)
	pack_output((ftmext_t *)self);
  
  FTMEXT_INLET_RETURN;  
}
*/

/****************************************************
 *
 *  initial arguments
 *
 */
static int
get_size_from_args(pack_t *self, int ac, const fts_atom_t *at)
{
  int size = ac;
  
  if(ac == 0)
	size = 2;
  if(ac == 1 && fts_is_number(at))
  {
	size = fts_get_number_int(at);
	
	if(size < 1)
	  size = 1;
  }
  
  return size;
}

FTMEXT_METHOD_VARARGS(pack, check_args)
{
  pack_t *self = (pack_t *)FTMEXT_GET_EXT();
  int ac = FTMEXT_GET_COUNT();
  fts_atom_t *at = FTMEXT_GET_ARGS();
  int size = get_size_from_args(self, ac, at);
  
  if(size != self->size)
	ftmext_error((ftmext_t *)self, "cannot accept expressions in initial arguments");
  
  FTMEXT_METHOD_RETURN;
}

/****************************************************
 *
 *  class
 *
 */
FTMEXT_INIT(pack)
{
  pack_t *self = (pack_t *)FTMEXT_GET_EXT();
  int ac = FTMEXT_GET_COUNT();
  fts_atom_t *at = FTMEXT_GET_ARGS();
  int size = get_size_from_args(self, ac, at);
  
  ftmext_set_inlets((ftmext_t *)self, size);
  
  if(ac <= 1)
  {
	int i;
	
	fts_array_init(&self->array, 0, NULL);
	
	for(i=0; i<size; i++)
	  fts_array_append_int(&self->array, 0);
  }
  else
	fts_array_init(&self->array, ac, at);	
  
  self->size = size;
	
  FTMEXT_INIT_RETURN;
}

FTMEXT_DELETE(pack)
{
  pack_t *self = (pack_t *)FTMEXT_GET_EXT();
  
  fts_array_destroy(&self->array);
  
  FTMEXT_DELETE_RETURN;
}

/* main function for Max/MSP */
FTMEXT_CLASS(fff, pack, pack_t)
{   
  FTMEXT_CLASS_DESCR(_prefix_ extclass, "FTM external example", "Example of an FTM external implementing an extended Max/MSP pack");
  
  FTMEXT_CLASS_ARGUMENTS_VARARGS(pack, check_args, NULL);
  FTMEXT_CLASS_ATTRIBUTE_VARARGS(pack, set, FTMEXT_ATTRIBUTE_DYNAMIC, "[<any: values> ...] - set values");
  
  FTMEXT_CLASS_MESSAGE_BANG(pack, output, "- output list stored values");
  
  FTMEXT_CLASS_INLET_ANY(pack, 0, "[<any: values> ...] - set values");
  FTMEXT_CLASS_INLET_MULTI(pack);
  
  /* alternative implementation */
  /* FTMEXT_CLASS_INLET_VARARGS(pack, 0, "[<any: an arbitrary value> - set the arbitrary value"); */
  
  FTMEXT_CLASS_OUTLET_VARARGS(pack, 0, "output list of stored values");
  
  FTMEXT_CLASS_RETURN;
}
